from .config import QUEUE_NAVIGATOR, QUEUE_OP


def front():
    """Go to the front of the queue."""
    with QUEUE_OP(sync_time="none", sync_remote=False):
        QUEUE_NAVIGATOR.go_to_front()


def before():
    """Go to the item before the current one."""
    with QUEUE_OP(sync_time="none", sync_remote=False):
        QUEUE_NAVIGATOR.before()


def after():
    """Go to the item after the current one."""
    with QUEUE_OP(sync_time="none", sync_remote=False):
        QUEUE_NAVIGATOR.after()


def back():
    """Go to the back of the queue."""
    with QUEUE_OP(sync_time="none", sync_remote=False):
        QUEUE_NAVIGATOR.go_to_back()


def status():
    """Print the current queue status."""
    with QUEUE_OP(sync_time="none", sync_remote=False):
        QUEUE_NAVIGATOR.status()



import enum
from typing import Annotated

import typer


class Location(str, enum.Enum):
    after = "after"
    back = "back"
    before = "before"
    front = "front"


LocationCLIOption = Annotated[Location, typer.Argument(help="Where to locate the new item.")]



from dataclasses import dataclass
from types import TracebackType
from typing import Literal

from rich import print

from ..queue.manipulator import QueueManipulator
from ..queue.navigator import QueueNavigator


@dataclass
class QueueOperation:
    queue_manipulator: QueueManipulator
    queue_navigator: QueueNavigator
    sync_time: Literal["enter", "exit", "none"] = "enter"
    sync_remote: bool = True

    def __call__(
        self, sync_remote: bool, sync_time: Literal["enter", "exit", "none"] = "enter"
    ) -> "QueueOperation":
        self.sync_remote = sync_remote
        self.sync_time = sync_time
        return self

    def __enter__(self):
        print(":arrows_clockwise: [bold green]Syncing with remote...[/bold green]")
        if self.sync_time == "enter":
            self.queue_manipulator.sync(sync_remote=self.sync_remote)

    def __exit__(self, exc_type: type, exc_val: Exception, exc_tb: TracebackType) -> None:
        if self.sync_time == "exit":
            self.queue_manipulator.sync(sync_remote=self.sync_remote)
        self.queue_navigator.status()



from dataclasses import dataclass
from typing import Protocol

from rich import box, print
from rich.panel import Panel

from ..cli.subprocess_utils import interactive_cmd, shell_output
from ..git import StagingMigrater


class QueueNavigator(Protocol):
    def go_to_front(self):
        ...

    def go_to_second_in_line(self):
        ...

    def go_to_back(self):
        ...

    def go_to_next_to_last(self):
        ...

    def before(self):
        ...

    def after(self):
        ...

    def status(self):
        ...


@dataclass(frozen=True)
class GraphiteNavigator(QueueNavigator):
    def go_to_front(self):
        with StagingMigrater():
            interactive_cmd("gt trunk")

    def go_to_second_in_line(self):
        self.go_to_front()
        self.after()

    def go_to_back(self):
        with StagingMigrater():
            interactive_cmd("gt top")

    def go_to_next_to_last(self):
        self.go_to_back()
        self.before()

    def before(self):
        interactive_cmd("gt down")

    def after(self):
        interactive_cmd("gt up")

    def status(self):
        result: str = shell_output("gt log short")  # type: ignore
        print(
            "\n",
            Panel(result, title="Back", subtitle="Front", expand=False, box=box.HORIZONTALS),
            "\n",
        )



from rich import print

from .config import ISSUE_CONTROLLER, QUEUE_MANIPULATOR, QUEUE_NAVIGATOR, QUEUE_OP
from .location import Location, LocationCLIOption


def add(location: LocationCLIOption = Location.after):
    """Add a new item to the current queue. Defaults to adding an item in between the current item and the next item."""
    with QUEUE_OP(sync_time="exit", sync_remote=False):
        selected_issue = ISSUE_CONTROLLER.select_issue()

        if location == Location.front:
            QUEUE_NAVIGATOR.go_to_front()
        elif location == Location.back:
            QUEUE_NAVIGATOR.go_to_back()
        elif location == Location.after:
            pass
        elif location == Location.before:
            QUEUE_NAVIGATOR.before()

        QUEUE_MANIPULATOR.add(selected_issue)
        ISSUE_CONTROLLER.label_issue_in_progress(selected_issue)


def move():
    """Move the current item to a new location in the queue."""
    with QUEUE_OP(sync_time="exit", sync_remote=False):
        QUEUE_MANIPULATOR.move()


def delete():
    """Delete a selected item from the queue."""
    with QUEUE_OP(sync_time="exit", sync_remote=False):
        QUEUE_MANIPULATOR.delete()


def fork(location: LocationCLIOption = Location.front):
    """Fork into a new queue and add an item. Defaults to forking from the first item in the current queue."""
    with QUEUE_OP(sync_time="enter", sync_remote=False):
        selected_issue = ISSUE_CONTROLLER.select_issue()

        if location == Location.front:
            QUEUE_NAVIGATOR.go_to_second_in_line()
        elif location == Location.back:
            QUEUE_NAVIGATOR.go_to_next_to_last()
        elif location == Location.after:
            pass  # No need to do anything, already in the correct location
        elif location == Location.before:
            QUEUE_NAVIGATOR.before()

        QUEUE_MANIPULATOR.fork(selected_issue)
        ISSUE_CONTROLLER.label_issue_in_progress(selected_issue)


def new():
    """Start a new queue on top of trunk."""
    with QUEUE_OP(sync_time="exit", sync_remote=False):
        selected_issue = ISSUE_CONTROLLER.select_issue()
        QUEUE_NAVIGATOR.go_to_front()
        QUEUE_MANIPULATOR.fork(selected_issue)
        ISSUE_CONTROLLER.label_issue_in_progress(selected_issue)


def sync(automerge: bool = False, draft: bool = False, squash: bool = False):
    """Synchronize all state, ensuring the queue is internally in sync, and in sync with the remote. Creates PRs if needed."""
    with QUEUE_OP(sync_time="none", sync_remote=False):
        QUEUE_MANIPULATOR.sync(automerge=automerge, squash=squash, draft=draft, sync_remote=True)
        print(":rocket: [bold green]Stack synced![/bold green]")
